/* -*- C++ -*-
 *
 * src/lemon/concept/undir_graph_component.h - Part of LEMON, a generic
 * C++ optimization library
 *
 * Copyright (C) 2004 Egervary Jeno Kombinatorikus Optimalizalasi
 * Kutatocsoport (Egervary Combinatorial Optimization Research Group,
 * EGRES).
 *
 * Permission to use, modify and distribute this software is granted
 * provided that this copyright notice appears in all copies. For
 * precise terms see the accompanying LICENSE file.
 *
 * This software is provided "AS IS" with no warranty of any kind,
 * express or implied, and with no claim as to its suitability for any
 * purpose.
 *
 */

///\ingroup graph_concepts
///\file
///\brief Undirected graphs and components of.


#ifndef LEMON_CONCEPT_UNDIR_GRAPH_H
#define LEMON_CONCEPT_UNDIR_GRAPH_H

#include <lemon/concept/graph_component.h>

namespace lemon {
  namespace concept {

    /// \addtogroup graph_concepts
    /// @{


    /// Skeleton class which describes an edge with direction in \ref
    /// UndirGraph "undirected graph".
    template <typename UndirEdge>
    class UndirGraphEdge : public UndirEdge {
    public:

      /// \e
      UndirGraphEdge() {}

      /// \e
      UndirGraphEdge(const UndirGraphEdge&) {}

      /// \e
      UndirGraphEdge(Invalid) {}

      /// \brief Constructs a directed version of an undirected edge
      ///
      /// \param forward If \c true the direction of the contructed edge
      /// is the same as the inherent direction of the \c undir_edge; if
      /// \c false --- the opposite.
      UndirGraphEdge(UndirEdge undir_edge, bool forward) {
	ignore_unused_variable_warning(undir_edge);
	ignore_unused_variable_warning(forward);
      }

      /// \e
      UndirGraphEdge& operator=(UndirGraphEdge) { return *this; }

      /// \e
      bool operator==(UndirGraphEdge) const { return true; }
      /// \e
      bool operator!=(UndirGraphEdge) const { return false; }

      /// \e
      bool operator<(UndirGraphEdge) const { return false; }

      template <typename Edge>
      struct Constraints {
	void constraints() {
	  /// \bug This should be is_base_and_derived ...
	  UndirEdge ue = e;
	  ue = e;
	  Edge forward(ue, true);
	  Edge backward(ue, false);

	  ignore_unused_variable_warning(forward);
	  ignore_unused_variable_warning(backward);
	}
	Edge e;
      };
    };
    

    struct BaseIterableUndirGraphConcept {

      template <typename Graph>
      struct Constraints {

	typedef typename Graph::UndirEdge UndirEdge;
	typedef typename Graph::Edge Edge;
	typedef typename Graph::Node Node;

	void constraints() {
	  checkConcept<BaseIterableGraphComponent, Graph>();
	  checkConcept<GraphItem<>, UndirEdge>();
	  checkConcept<UndirGraphEdge<UndirEdge>, Edge>();

	  graph.first(ue);
	  graph.next(ue);

	  const_constraints();
	}
	void const_constraints() {
	  Node n;
	  n = graph.target(ue);
	  n = graph.source(ue);
	  n = graph.oppositeNode(n0, ue);

	  bool b;
	  b = graph.forward(e);
	  ignore_unused_variable_warning(b);
	}

	Graph graph;
	Edge e;
	Node n0;
	UndirEdge ue;
      };

    };


    struct IterableUndirGraphConcept {

      template <typename Graph>
      struct Constraints {
	void constraints() {
	  /// \todo we don't need the iterable component to be base iterable
	  /// Don't we really???
	  //checkConcept< BaseIterableUndirGraphConcept, Graph > ();

	  checkConcept<IterableGraphComponent, Graph> ();

	  typedef typename Graph::UndirEdge UndirEdge;
	  typedef typename Graph::UndirEdgeIt UndirEdgeIt;
	  typedef typename Graph::IncEdgeIt IncEdgeIt;

	  checkConcept<GraphIterator<Graph, UndirEdge>, UndirEdgeIt>();
	  checkConcept<GraphIncIterator<Graph, UndirEdge>, IncEdgeIt>();
	}
      };

    };

    struct MappableUndirGraphConcept {

      template <typename Graph>
      struct Constraints {

	struct Dummy {
	  int value;
	  Dummy() : value(0) {}
	  Dummy(int _v) : value(_v) {}
	};

	void constraints() {
	  checkConcept<MappableGraphComponent, Graph>();

	  typedef typename Graph::template UndirEdgeMap<int> IntMap;
	  checkConcept<GraphMap<Graph, typename Graph::UndirEdge, int>,
	    IntMap >();

	  typedef typename Graph::template UndirEdgeMap<bool> BoolMap;
	  checkConcept<GraphMap<Graph, typename Graph::UndirEdge, bool>,
	    BoolMap >();

	  typedef typename Graph::template UndirEdgeMap<Dummy> DummyMap;
	  checkConcept<GraphMap<Graph, typename Graph::UndirEdge, Dummy>,
	    DummyMap >();
	}
      };

    };

    struct ExtendableUndirGraphConcept {

      template <typename Graph>
      struct Constraints {
	void constraints() {
	  node_a = graph.addNode();
	  uedge = graph.addEdge(node_a, node_b);
	}
	typename Graph::Node node_a, node_b;
	typename Graph::UndirEdge uedge;
	Graph graph;
      };

    };

    struct ErasableUndirGraphConcept {

      template <typename Graph>
      struct Constraints {
	void constraints() {
	  graph.erase(n);
	  graph.erase(e);
	}
	Graph graph;
	typename Graph::Node n;
	typename Graph::UndirEdge e;
      };

    };

    /// Class describing the concept of Undirected Graphs.

    /// This class describes the common interface of all Undirected
    /// Graphs.
    ///
    /// As all concept describing classes it provides only interface
    /// without any sensible implementation. So any algorithm for
    /// undirected graph should compile with this class, but it will not
    /// run properly, of couse.
    ///
    /// In LEMON undirected graphs also fulfill the concept of directed
    /// graphs (\ref lemon::concept::Graph "Graph Concept"). For
    /// explanation of this and more see also the page \ref undir_graphs,
    /// a tutorial about undirected graphs.

    class UndirGraph {
    public:

      /// Type describing a node in the graph
      typedef GraphNode Node;

      /// Type describing an undirected edge
      typedef GraphItem<'u'> UndirEdge;

      /// Type describing an UndirEdge with direction
#ifndef DOXYGEN
      typedef UndirGraphEdge<UndirEdge> Edge;
#else
      typedef UndirGraphEdge Edge;
#endif

      /// Iterator type which iterates over all nodes
#ifndef DOXYGEN
      typedef GraphIterator<UndirGraph, Node> NodeIt;
#else
      typedef GraphIterator NodeIt;
#endif

      /// Iterator type which iterates over all undirected edges
#ifndef DOXYGEN
      typedef GraphIterator<UndirGraph, UndirEdge> UndirEdgeIt;
#else
      typedef GraphIterator UndirEdgeIt;
#endif

      /// Iterator type which iterates over all directed edges.

      /// Iterator type which iterates over all edges (each undirected
      /// edge occurs twice with both directions.
#ifndef DOXYGEN
      typedef GraphIterator<UndirGraph, Edge> EdgeIt;
#else
      typedef GraphIterator EdgeIt;
#endif


      /// Iterator of undirected edges incident to a node
#ifndef DOXYGEN
      typedef GraphIncIterator<UndirGraph, UndirEdge, 'u'> IncEdgeIt;
#else
      typedef GraphIncIterator IncEdgeIt;
#endif

      /// Iterator of edges incoming to a node
#ifndef DOXYGEN
      typedef GraphIncIterator<UndirGraph, Edge, 'i'> InEdgeIt;
#else
      typedef GraphIncIterator InEdgeIt;
#endif

      /// Iterator of edges outgoing from a node
#ifndef DOXYGEN
      typedef GraphIncIterator<UndirGraph, Edge, 'o'> OutEdgeIt;
#else
      typedef GraphIncIterator OutEdgeIt;
#endif

      /// NodeMap template
#ifdef DOXYGEN
      typedef GraphMap NodeMap<T>;
#endif

      /// UndirEdgeMap template
#ifdef DOXYGEN
      typedef GraphMap UndirEdgeMap<T>;
#endif

      /// EdgeMap template
#ifdef DOXYGEN
      typedef GraphMap EdgeMap<T>;
#endif

      template <typename T>
      class NodeMap : public GraphMap<UndirGraph, Node, T> {
	typedef GraphMap<UndirGraph, Node, T> Parent;
      public:

	explicit NodeMap(const UndirGraph &g) : Parent(g) {}
	NodeMap(const UndirGraph &g, T t) : Parent(g, t) {}
      };

      template <typename T>
      class UndirEdgeMap : public GraphMap<UndirGraph, UndirEdge, T> {
	typedef GraphMap<UndirGraph, UndirEdge, T> Parent;
      public:

	explicit UndirEdgeMap(const UndirGraph &g) : Parent(g) {}
	UndirEdgeMap(const UndirGraph &g, T t) : Parent(g, t) {}
      };

      template <typename T>
      class EdgeMap : public GraphMap<UndirGraph, Edge, T> {
	typedef GraphMap<UndirGraph, Edge, T> Parent;
      public:

	explicit EdgeMap(const UndirGraph &g) : Parent(g) {}
	EdgeMap(const UndirGraph &g, T t) : Parent(g, t) {}
      };

      /// Is the Edge oriented "forward"?

      /// Returns whether the given directed edge is same orientation as
      /// the corresponding undirected edge.
      ///
      /// \todo "What does the direction of an undirected edge mean?"
      bool forward(Edge) const { return true; }

      /// Opposite node on an edge

      /// \return the opposite of the given Node on the given Edge
      ///
      /// \todo What should we do if given Node and Edge are not incident?
      Node oppositeNode(Node, UndirEdge) const { return INVALID; }

      /// First node of the undirected edge.

      /// \return the first node of the given UndirEdge.
      ///
      /// Naturally undirectected edges don't have direction and thus
      /// don't have source and target node. But we use these two methods
      /// to query the two endnodes of the edge. The direction of the edge
      /// which arises this way is called the inherent direction of the
      /// undirected edge, and is used to define the "forward" direction
      /// of the directed versions of the edges.
      /// \sa forward
      Node source(UndirEdge) const { return INVALID; }

      /// Second node of the undirected edge.
      Node target(UndirEdge) const { return INVALID; }

      /// Source node of the directed edge.
      Node source(Edge) const { return INVALID; }

      /// Target node of the directed edge.
      Node target(Edge) const { return INVALID; }

      /// First node of the graph

      /// \note This method is part of so called \ref
      /// developpers_interface "Developpers' interface", so it shouldn't
      /// be used in an end-user program.
      void first(Node&) const {}
      /// Next node of the graph

      /// \note This method is part of so called \ref
      /// developpers_interface "Developpers' interface", so it shouldn't
      /// be used in an end-user program.
      void next(Node&) const {}

      /// First undirected edge of the graph

      /// \note This method is part of so called \ref
      /// developpers_interface "Developpers' interface", so it shouldn't
      /// be used in an end-user program.
      void first(UndirEdge&) const {}
      /// Next undirected edge of the graph

      /// \note This method is part of so called \ref
      /// developpers_interface "Developpers' interface", so it shouldn't
      /// be used in an end-user program.
      void next(UndirEdge&) const {}

      /// First directed edge of the graph

      /// \note This method is part of so called \ref
      /// developpers_interface "Developpers' interface", so it shouldn't
      /// be used in an end-user program.
      void first(Edge&) const {}
      /// Next directed edge of the graph

      /// \note This method is part of so called \ref
      /// developpers_interface "Developpers' interface", so it shouldn't
      /// be used in an end-user program.
      void next(Edge&) const {}

      /// First outgoing edge from a given node

      /// \note This method is part of so called \ref
      /// developpers_interface "Developpers' interface", so it shouldn't
      /// be used in an end-user program.
      void firstOut(Edge&, Node) const {}
      /// Next outgoing edge to a node

      /// \note This method is part of so called \ref
      /// developpers_interface "Developpers' interface", so it shouldn't
      /// be used in an end-user program.
      void nextOut(Edge&) const {}

      /// First incoming edge to a given node

      /// \note This method is part of so called \ref
      /// developpers_interface "Developpers' interface", so it shouldn't
      /// be used in an end-user program.
      void firstIn(Edge&, Node) const {}
      /// Next incoming edge to a node

      /// \note This method is part of so called \ref
      /// developpers_interface "Developpers' interface", so it shouldn't
      /// be used in an end-user program.
      void nextIn(Edge&) const {}


      template <typename Graph>
      struct Constraints {
	void constraints() {
	  checkConcept<BaseIterableUndirGraphConcept, Graph>();
	  checkConcept<IterableUndirGraphConcept, Graph>();
	  checkConcept<MappableUndirGraphConcept, Graph>();
	}
      };

    };

    class ExtendableUndirGraph : public UndirGraph {
    public:

      template <typename Graph>
      struct Constraints {
	void constraints() {
	  checkConcept<BaseIterableUndirGraphConcept, Graph>();
	  checkConcept<IterableUndirGraphConcept, Graph>();
	  checkConcept<MappableUndirGraphConcept, Graph>();

	  checkConcept<UndirGraph, Graph>();
	  checkConcept<ExtendableUndirGraphConcept, Graph>();
	  checkConcept<ClearableGraphComponent, Graph>();
	}
      };

    };

    class ErasableUndirGraph : public ExtendableUndirGraph {
    public:

      template <typename Graph>
      struct Constraints {
	void constraints() {
	  checkConcept<ExtendableUndirGraph, Graph>();
	  checkConcept<ErasableUndirGraphConcept, Graph>();
	}
      };

    };

    /// @}

  }

}

#endif
